#####
##### Signal extrema specification
#####

struct SignalExtrema
    physical_min::Float32
    physical_max::Float32
    digital_min::Float32
    digital_max::Float32
end

function SignalExtrema(signal::Onda.Signal)
    digital_extrema = (typemin(signal.sample_type), typemax(signal.sample_type))
    physical_extrema = @. (signal.sample_resolution_in_unit * digital_extrema) + signal.sample_offset_in_unit
    return SignalExtrema(physical_extrema..., digital_extrema...)
end

#####
##### `EDF.FileHeader` conversion
#####

const DATA_RECORD_SIZE_LIMIT = 30720
const EDF_BYTE_LIMIT = 8

edf_sample_count_per_record(signal::Onda.Signal, seconds_per_record::Float64) = Int16(signal.sample_rate * seconds_per_record)

function edf_record_metadata(signals::Vector{Onda.Signal})
    sample_rates = map(signal -> rationalize(signal.sample_rate), signals)
    seconds_per_record = lcm(map(denominator, sample_rates))
    samples_per_record = map(zip(signals, sample_rates)) do (signal, sample_rate)
        return channel_count(signal) * numerator(sample_rate) * seconds_per_record
    end
    if sum(samples_per_record) > DATA_RECORD_SIZE_LIMIT
        if length(signals) == 1
            seconds_per_record = Float64(inv(first(sample_rates)))
        else
            scale = gcd(numerator.(sample_rates) .* seconds_per_record)
            samples_per_record ./= scale
            sum(samples_per_record) > DATA_RECORD_SIZE_LIMIT && throw(RecordSizeException(signals))
            seconds_per_record /= scale
        end
        sizeof(string(seconds_per_record)) > EDF_BYTE_LIMIT && throw(EDFPrecisionError(seconds_per_record))
    end
    record_duration_in_nanoseconds = Nanosecond(seconds_per_record * 1_000_000_000)
    signal_duration = maximum(signal -> signal.stop_nanosecond, signals)
    record_count = ceil(signal_duration / record_duration_in_nanoseconds)
    sizeof(string(record_count)) > EDF_BYTE_LIMIT && throw(EDFPrecisionError(record_count))
    return record_count, seconds_per_record
end

struct RecordSizeException <: Exception
    signals::Vector{Onda.Signal}
end

struct EDFPrecisionError <: Exception
    value::Number
end

function Base.showerror(io::IO, exception::RecordSizeException)
    print(io, "RecordSizeException: sample rates ")
    print(io, [signal.sample_rate for signal in exception.signals])
    print(io, " cannot be resolved to a data record size smaller than ")
    print(io, DATA_RECORD_SIZE_LIMIT * 2, " bytes")
end

function Base.showerror(io::IO, exception::EDFPrecisionError)
    print(io, "EDFPrecisionError: String representation of value ")
    print(io, exception.value)
    print(io, " is longer than 8 ASCII characters")
end

#####
##### `EDF.Signal` conversion
#####

function export_edf_label(signal_name::Symbol, channel_name::Symbol)
    signal_edf_name = uppercase(string(signal_name))
    channel_edf_name = uppercase(string(channel_name))
    if signal_edf_name == "EEG" && !('-' in channel_edf_name)
        channel_edf_name *= "-Ref"
    end
    return string(signal_edf_name, " ", channel_edf_name)
end

function export_edf_header(signals::Vector{Onda.Signal};
                           version::AbstractString="0",
                           patient_metadata=EDF.PatientID(missing, missing, missing, missing),
                           recording_metadata=EDF.RecordingID(missing, missing, missing, missing),
                           is_contiguous::Bool=true,
                           start::DateTime=DateTime(Year(1985)))
    return EDF.FileHeader(version, patient_metadata, recording_metadata, start,
                          is_contiguous, edf_record_metadata(signals)...)
end

function export_edf_signals(dataset::Dataset, uuid::UUID, signals, seconds_per_record::Float64)
    edf_signals = Union{EDF.AnnotationsSignal,EDF.Signal}[]
    for (signal_name::Symbol, onda_signal::Onda.Signal) in signals
        if sizeof(onda_signal.sample_type) > sizeof(Int16)
            decoded_samples = Onda.load(dataset, uuid, signal_name)
            scaled_resolution = onda_signal.sample_resolution_in_unit * (sizeof(onda_signal.sample_type) / sizeof(Int16))
            signal = signal_from_template(onda_signal; sample_type=Int16, sample_resolution_in_unit=scaled_resolution)
            samples = encode(Onda.Samples(signal, false, decoded_samples.data), missing)
        else
            signal = onda_signal
            samples = Onda.load_encoded(dataset, uuid, signal_name)
        end
        extrema = SignalExtrema(signal)
        for channel_name in onda_signal.channel_names
            sample_count = edf_sample_count_per_record(onda_signal, seconds_per_record)
            physical_dimension = onda_to_edf_unit(onda_signal.sample_unit)
            edf_signal_header = EDF.SignalHeader(export_edf_label(signal_name, channel_name),
                                                 "", physical_dimension,
                                                 extrema.physical_min, extrema.physical_max,
                                                 extrema.digital_min, extrema.digital_max,
                                                 "", sample_count)
            sample_data = vec(samples[channel_name, :].data)
            padding = Iterators.repeated(zero(Int16), (sample_count - (length(sample_data) % sample_count)) % sample_count)
            edf_signal_samples = append!(sample_data, padding)
            push!(edf_signals, EDF.Signal(edf_signal_header, edf_signal_samples))
        end
    end
    return edf_signals
end

#####
##### `export_edf`
#####

"""
    export_edf(dataset::Dataset, uuid::UUID; signal_names=keys(dataset.recordings[uuid].signals))

Return an `EDF.File` containing signal data converted from `dataset.recordings[uuid]`.

Each `EDF.Signal` in the returned `EDF.File` corresponds to a channel of an `Onda.Signal`;
only signals whose names are given in `signal_name`s are exported.

The ordering of `EDF.Signal`s in the output will match the order of the given `signal_names`
(and within each channel grouping, the order of the signal's channels).
"""
function export_edf(dataset::Dataset, uuid::UUID;
                    signal_names=keys(dataset.recordings[uuid].signals),
                    export_annotations::Bool=true,
                    kwargs...)
    recording = dataset.recordings[uuid]
    signals = [recording.signals[name] for name in signal_names]
    edf_header = export_edf_header(signals; kwargs...)
    edf_signals = export_edf_signals(dataset, uuid, zip(signal_names, signals),
                                     edf_header.seconds_per_record)
    if export_annotations
        records = [[EDF.TimestampedAnnotationList(edf_header.seconds_per_record * i, nothing, String[""])]
                   for i in 0:(edf_header.record_count - 1)]
        for annotation in sort(collect(recording.annotations); by=first)
            annotation_onset_in_seconds = first(annotation).value / 1e9
            annotation_duration_in_seconds = duration(annotation).value / 1e9
            matching_record = records[Int(fld(annotation_onset_in_seconds, edf_header.seconds_per_record)) + 1]
            tal = EDF.TimestampedAnnotationList(annotation_onset_in_seconds, annotation_duration_in_seconds, [annotation.value])
            push!(matching_record, tal)
        end
        push!(edf_signals, EDF.AnnotationsSignal(records))
    end
    return EDF.File((io = IOBuffer(); close(io); io), edf_header, edf_signals)
end
